/*
 * $Id: CloneableTester.java,v 1.8 2011/03/22 18:22:59 oboehm Exp $
 *
 * Copyright (c) 2010 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * (c)reated 06.08.2010 by oliver (ob@oasd.de)
 */

package patterntesting.runtime.junit;

import java.lang.reflect.*;
import java.util.*;

import org.apache.commons.logging.*;

import patterntesting.runtime.monitor.ClasspathMonitor;

/**
 * This tester checks class which implements Clonable and has therefore
 * the clone method implemented. According {@link Object#clone()} the
 * following conditions should be true:
 * <blockquote>
 * <pre>
 * x.clone() != x</pre></blockquote>
 * will be true, and that the expression:
 * <blockquote>
 * <pre>
 * x.clone().getClass() == x.getClass()</pre></blockquote>
 * will be <tt>true</tt>, but these are not absolute requirements.
 * While it is typically the case that:
 * <blockquote>
 * <pre>
 * x.clone().equals(x)</pre></blockquote>
 * will be <tt>true</tt>, this is not an absolute requirement.
 * <p>
 * So the equals method will be only checked if it is overwritten.
 * <br/>
 * NOTE: In the future this class will be perhaps part of the ObjectTester
 * class.
 * <br/>
 * Before v1.1 the methods are named "checkCloning". Since 1.1 these
 * methods will have now an "assert" prefix ("assertCloning").
 * 
 * TODO: remove deprecated check methods in 1.3
 *
 * @author oliver
 * @since 1.0.2 (06.08.2010)
 */
public final class CloneableTester {

    private static final Log log = LogFactory.getLog(CloneableTester.class);
    private static final ClasspathMonitor classpathMonitor = ClasspathMonitor.getInstance();

    /** Utility class - no need to instantiate it. */
    private CloneableTester() {}

    /**
     * Check cloning.
     *
     * @param clazz the clazz
     * @since 1.1
     */
    public static void assertCloning(final Class<? extends Cloneable> clazz) {
        try {
            assertCloning(clazz.newInstance());
        } catch (InstantiationException e) {
            throw new AssertionError("can't instantiate " + clazz + ":" + e);
        } catch (IllegalAccessException e) {
            throw new AssertionError("can't access " + clazz + ":" + e);
        }
    }

    /**
     * Check cloning.
     *
     * @param clazz the clazz
     * @deprecated use {@link #assertCloning(Class)}
     */
    @Deprecated
    public static void checkCloning(final Class<? extends Cloneable> clazz) {
        assertCloning(clazz);
    }

    /**
     * We call the clone method of the given orig paramter.Because the
     * clone method is normally "protected" we use reflection to call it.
     * Then we compare the orig and cloned object which should be equals.
     *
     * @param orig the original object
     * @since 1.1
     */
    public static void assertCloning(final Cloneable orig) {
        Cloneable clone = getCloneOf(orig);
        if (ObjectTester.hasEqualsDeclared(orig.getClass())) {
            ObjectTester.assertEquals(orig, clone);
        }
    }
    
    /**
     * We call the clone method of the given orig paramter.Because the
     * clone method is normally "protected" we use reflection to call it.
     * Then we compare the orig and cloned object which should be equals.
     *
     * @param orig the original object
     * @deprecated use {@link #assertCloning(Cloneable)}
     */
    @Deprecated
    public static void checkCloning(final Cloneable orig) {
        assertCloning(orig);
    }

    /**
     * Gets the clone of the given Cloneable object.
     *
     * @param orig the orig
     * @return the clone of
     * @throws AssertionError the assertion error
     */
    public static Cloneable getCloneOf(final Cloneable orig) throws AssertionError {
        try {
            Method cloneMethod = getPublicCloneMethod(orig.getClass());
            Cloneable clone = (Cloneable) cloneMethod.invoke(orig);
            if (clone == orig) {
                throw new AssertionError(clone + " must have another reference as original object");
            }
            return clone;
        } catch (SecurityException e) {
            throw new AssertionError("can't access clone method of " + orig.getClass() + ": " + e);
        } catch (IllegalAccessException e) {
            throw new AssertionError("can't access clone method of " + orig.getClass() + ": " + e);
        } catch (InvocationTargetException e) {
            throw new AssertionError("clone of " + orig.getClass() + " failed: " + e);
        }
    }

    private static Method getPublicCloneMethod(final Class<? extends Cloneable> cloneClass) throws AssertionError {
        try {
            return cloneClass.getMethod("clone");
        } catch (NoSuchMethodException nsme) {
            if (hasCloneMethod(cloneClass)) {
                throw new AssertionError("clone() is not public in " + cloneClass);
            }
            AssertionError error = new AssertionError("no clone method found in " + cloneClass);
            error.initCause(nsme);
            throw error;
        }
    }

    private static boolean hasCloneMethod(final Class<? extends Cloneable> cloneClass) {
        try {
            Method m = cloneClass.getDeclaredMethod("clone");
            if (log.isTraceEnabled()) {
                log.trace(m + " found in " + cloneClass);
            }
            return true;
        } catch (NoSuchMethodException e) {
            return false;
        }
    }

    /**
     * Check for each class in the given collection if it can be cloned
     * correct.
     *
     * @param classes a collection of classes to be checked
     * @since 1.1
     */
    public static void assertCloning(final Collection<Class<? extends Cloneable>> classes) {
        for (Class<? extends Cloneable> clazz : classes) {
            assertCloning(clazz);
        }
    }

    /**
     * Check for each class in the given collection if it can be cloned
     * correct.
     *
     * @param classes a collection of classes to be checked
     * @deprecated use {@link #assertCloning(Collection)}
     */
    @Deprecated
    public static void checkCloning(final Collection<Class<? extends Cloneable>> classes) {
        assertCloning(classes);
    }

    /**
     * Check for each class in the given package if it can be cloned
     * correct.
     * <br/>
     * To get a name of a package call {@link Package#getPackage(String)}.
     * But be sure that you can't get <em>null</em> as result. In this case
     * use {@link #assertCloningOfPackage(String)}.
     *
     * @param pkg the package e.g. "patterntesting.runtime"
     * @see #assertCloningOfPackage(String)
     * @since 1.1
     */
    public static void assertCloning(final Package pkg) {
        assert pkg!= null;
        assertCloningOfPackage(pkg.getName());
    }
    
    /**
     * Check for each class in the given package if it can be cloned
     * correct.
     * <br/>
     * To get a name of a package call {@link Package#getPackage(String)}.
     * But be sure that you can't get <em>null</em> as result. In this case
     * use {@link #assertCloningOfPackage(String)}.
     *
     * @param pkg the package e.g. "patterntesting.runtime"
     * @deprecated use {@link #assertCloning(Package)}
     */
    @Deprecated
    public static void checkCloning(final Package pkg) {
        assertCloning(pkg);
    }

    /**
     * Check for each class in the given package if it can be cloned
     * correct.
     *
     * @param packageName the package name e.g. "patterntesting.runtime"
     * @since 1.1
     */
    public static void assertCloningOfPackage(final String packageName) {
        assert packageName != null;
        Collection<Class<? extends Cloneable>> cloneables = getCloneableClasses(packageName);
        assertCloning(cloneables);
    }

    /**
     * Check for each class in the given package if it can be cloned
     * correct.
     *
     * @param packageName the package name e.g. "patterntesting.runtime"
     * @deprecated use {@link #assertCloningOfPackage(String)}
     */
    @Deprecated
    public static void checkCloningOfPackage(final String packageName) {
        assertCloningOfPackage(packageName);
    }

    /**
     * Check for each class in the given package if the clone method is
     * implemented correct.
     *
     * @param packageName the package name e.g. "patterntesting.runtime"
     * @param excluded classes which should be excluded from the check
     * @see #assertCloningOfPackage(String)
     * @since 1.1
     */
    public static void assertCloningOfPackage(final String packageName,
            final Class<? extends Cloneable>... excluded) {
        List<Class<? extends Cloneable>> excludedList = Arrays.asList(excluded);
        assertCloningOfPackage(packageName, excludedList);
    }

    /**
     * Check for each class in the given package if the clone method is
     * implemented correct.
     *
     * @param packageName the package name e.g. "patterntesting.runtime"
     * @param excluded classes which should be excluded from the check
     * @deprecated use {@link #assertCloningOfPackage(String, Class...)}
     */
    @Deprecated
    public static void checkCloningOfPackage(final String packageName,
            final Class<Cloneable>... excluded) {
        assertCloningOfPackage(packageName, excluded);
    }

    /**
     * Check for each class in the given package if the clone method is
     * implemented correct.
     *
     * @param packageName the package name e.g. "patterntesting.runtime"
     * @param excluded classes which should be excluded from the check
     * @see #assertCloningOfPackage(String)
     * @since 1.1
     */
    public static void assertCloningOfPackage(final String packageName,
            final List<Class<? extends Cloneable>> excluded) {
        Collection<Class<? extends Cloneable>> classes = getCloneableClasses(packageName);
        log.debug(excluded + " will be excluded from check");
        ObjectTester.removeClasses(classes, excluded);
        assertCloning(classes);
    }

    /**
     * Check for each class in the given package if the clone method is
     * implemented correct.
     *
     * @param packageName the package name e.g. "patterntesting.runtime"
     * @param excluded classes which should be excluded from the check
     * @deprecated use {@link #assertCloningOfPackage(String, List)}
     */
    @Deprecated
    public static void checkCloningOfPackage(final String packageName,
            final List<Class<? extends Cloneable>> excluded) {
        assertCloningOfPackage(packageName, excluded);
    }

    @SuppressWarnings("unchecked")
    private static Collection<Class<? extends Cloneable>> getCloneableClasses(
            final String packageName) {
        Collection<Class<? extends Cloneable>> cloneables = new ArrayList<Class<? extends Cloneable>>();
        Collection<Class<?>> concreteClasses = classpathMonitor.getConcreteClassList(packageName);
        for (Class<?> clazz : concreteClasses) {
            if (Cloneable.class.isAssignableFrom(clazz)) {
                Class<Cloneable> cloneClass = (Class<Cloneable>) clazz;
                cloneables.add(cloneClass);
                if (log.isTraceEnabled()) {
                    log.trace("Cloneable found: " + clazz);
                }
            }
        }
        return cloneables;
    }

}

