/*
 * (c)reated 21.07.2010 by oliver
 */
package patterntesting.runtime.junit;

import java.io.*;
import java.util.*;

import org.apache.commons.logging.*;

import patterntesting.runtime.monitor.ClasspathMonitor;
import patterntesting.runtime.util.Converter;

/**
 * This is a utility class to check the serializable nature of classes.
 * <br/>
 * NOTE: In the future this class will be perhaps part of the ObjectTester
 * class.
 * <br/>
 * Before v1.1 the methods are named "checkSerialization". Since 1.1 these
 * methods will have now an "assert" prefix ("assertSerialization").
 * 
 * TODO: remove deprecated check methods in 1.3
 *
 * @author oliver (oliver.boehm@agentes.de)
 * @since 1.0.3 (21.07.2010)
 */
public final class SerializableTester {

	private static final Log log = LogFactory.getLog(SerializableTester.class);
    private static final ClasspathMonitor classpathMonitor = ClasspathMonitor.getInstance();

	/** Utitlity class - no need to instantiate it. */
	private SerializableTester() {}

    /**
     * Check serialization and deserialization of an object.
     * Because not all classes has a (correct) overwritten equals method
     * we do not compare the deserialized object with the equals() method.
     * But if it implements Comparable we verify the equality of the two
     * objects with the compareTo() method.
     *
     * @param object the object
     * @throws NotSerializableException if object can't serialized
     * @since 1.1
     */
    public static void assertSerialization(final Serializable object)
            throws NotSerializableException {
    	byte[] bytes = Converter.serialize(object);
    	log.debug(object + " serialized in " + bytes.length + " bytes");
    	try {
            Object deserialized = Converter.deserialize(bytes);
            if (object instanceof Comparable<?>) {
                ObjectTester.assertCompareTo(object, deserialized);
            }
        } catch (ClassNotFoundException canthappen) {
            throw new RuntimeException(canthappen);
        }
    }

    /**
     * Check serialization and deserialization of an object.
     * Because not all classes has a (correct) overwritten equals method
     * we do not compare the deserialized object with the equals() method.
     * But if it implements Comparable we verify the equality of the two
     * objects with the compareTo() method.
     *
     * @param object the object
     * @throws NotSerializableException if object can't serialized
     * @deprecated use {@link #assertSerialization(Serializable)}
     */
    @Deprecated
    public static void checkSerialization(final Serializable object)
            throws NotSerializableException {
        assertSerialization(object);
    }

    /**
     * This method will create an object of the given class using the
     * default constructor. So three preconditions must be true:
     * <ol>
     *  <li>the class must not be abstract</li>
     *  <li>there must be a (public) default constructor</li>
     *  <li>it must be Serializable</li>
     * </ol>
     *
     * @param clazz the clazz
     * @throws NotSerializableException if the check fails
     * @since 1.1
     */
    public static void assertSerialization(final Class<Serializable> clazz) throws NotSerializableException {
        if (log.isTraceEnabled()) {
            log.trace("checking " + clazz.getName() + " if it can be serialized...");
        }
        try {
            Serializable obj = clazz.newInstance();
            assertSerialization(obj);
        } catch (InstantiationException e) {
            throw new IllegalArgumentException("can't instantiate " + clazz, e);
        } catch (IllegalAccessException e) {
            throw new IllegalArgumentException("can't access ctor of " + clazz, e);
        }
    }

    /**
     * This method will create an object of the given class using the
     * default constructor. So three preconditions must be true:
     * <ol>
     *  <li>the class must not be abstract</li>
     *  <li>there must be a (public) default constructor</li>
     *  <li>it must be Serializable</li>
     * </ol>
     *
     * @param clazz the clazz
     * @throws NotSerializableException if the check fails
     * @deprecated use {@link #assertSerialization(Serializable)}
     */
    @Deprecated
    public static void checkSerialization(final Class<Serializable> clazz) throws NotSerializableException {
        assertSerialization(clazz);
    }

    /**
     * Check for each class in the given collection if it can be serialized and
     * deserialized.
     *
     * @param classes a collection of classes to be checked
     * @throws NotSerializableException if one of the classes can't be serialized
     * @since 1.1
     */
    public static void assertSerialization(final Collection<Class<Serializable>> classes) throws NotSerializableException {
        for (Class<?> clazz : classes) {
            assertSerialization(clazz);
        }
    }

    /**
     * Check for each class in the given collection if it can be serialized and
     * deserialized.
     *
     * @param classes a collection of classes to be checked
     * @throws NotSerializableException if one of the classes can't be serialized
     * @deprecated use {@link #assertSerialization(Collection)}
     */
    @Deprecated
    public static void checkSerialization(final Collection<Class<Serializable>> classes) throws NotSerializableException {
        assertSerialization(classes);
    }

    /**
     * Check for each class in the given package if it can be serialized and
     * deserialized.
     * <br/>
     * To get a name of a package call {@link Package#getPackage(String)}.
     * But be sure that you can't get <em>null</em> as result. In this case
     * use {@link #assertSerializationOfPackage(String)}.
     *
     * @param pkg the package e.g. "patterntesting.runtime"
     * @throws NotSerializableException if one of the class can't be serialized
     * @see #assertSerializationOfPackage(String)
     * @since 1.1
     */
    public static void assertSerialization(final Package pkg) throws NotSerializableException {
        assert pkg!= null;
        assertSerializationOfPackage(pkg.getName());
    }

    /**
     * Check for each class in the given package if it can be serialized and
     * deserialized.
     * <br/>
     * To get a name of a package call {@link Package#getPackage(String)}.
     * But be sure that you can't get <em>null</em> as result. In this case
     * use {@link #assertSerializationOfPackage(String)}.
     *
     * @param pkg the package e.g. "patterntesting.runtime"
     * @throws NotSerializableException if one of the class can't be serialized
     * @see #assertSerializationOfPackage(String)
     * @deprecated use {@link #assertSerialization(Package)}
     */
    @Deprecated
    public static void checkSerialization(final Package pkg) throws NotSerializableException {
        assertSerialization(pkg);
    }

    /**
     * Check for each class in the given package if it can be serialized and
     * deserialized.
     *
     * @param packageName the package name e.g. "patterntesting.runtime"
     * @throws NotSerializableException if one of the class can't be serialized
     * @since 1.1
     */
    public static void assertSerializationOfPackage(final String packageName) throws NotSerializableException {
        assert packageName != null;
        Collection<Class<Serializable>> classes = getSerializableClasses(packageName);
        assertSerialization(classes);
    }

    /**
     * Check for each class in the given package if it can be serialized and
     * deserialized.
     *
     * @param packageName the package name e.g. "patterntesting.runtime"
     * @throws NotSerializableException if one of the class can't be serialized
     * @deprecated use {@link #assertSerializationOfPackage(String)}
     */
    @Deprecated
    public static void checkSerializationOfPackage(final String packageName) throws NotSerializableException {
        assertSerialization(packageName);
    }

    @SuppressWarnings("unchecked")
    private static Collection<Class<Serializable>> getSerializableClasses(final String packageName) {
        String[] classnames = classpathMonitor.getClasspathClasses();
        Collection<Class<Serializable>> classes = new ArrayList<Class<Serializable>>();
        for (int i = 0; i < classnames.length; i++) {
            if (classnames[i].startsWith(packageName)) {
                try {
                    Class<?> clazz = Class.forName(classnames[i]);
                    if (Serializable.class.isAssignableFrom(clazz)) {
                        classes.add((Class<Serializable>) clazz);
                    }
                } catch (ClassNotFoundException ignored) {
                    log.info(classnames[i] + " will be ignored", ignored);
                }
            }
        }
        return classes;
    }

}
