/*
 * $Id: SmokeFilter.java,v 1.13 2010/12/31 14:40:13 oboehm Exp $
 *
 * Copyright (c) 2010 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * (c)reated 03.05.2010 by oliver (ob@oasd.de)
 */

package patterntesting.runtime.junit;

import java.util.*;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.*;
import org.junit.runner.Description;
import org.junit.runner.manipulation.Filter;

import patterntesting.runtime.NullConstants;
import patterntesting.runtime.annotation.*;
import patterntesting.runtime.junit.internal.TestOn;
import patterntesting.runtime.util.*;

/**
 * This filter handles the different annotations like Broken, RunTestOn,
 * SkipTestOn, IntegrationTest or SmokeTest.
 *
 * @author oliver
 * @since 1.0 (03.05.2010)
 */
public final class SmokeFilter extends Filter {

    private static final Log log = LogFactory.getLog(SmokeFilter.class);
    private Date today = new Date();
    private String osName = Environment.OS_NAME;
    private String userName = Environment.USER_NAME;
    private int filteredOut = 0;
    private final Dictionary<Description, Boolean> shouldRunCache = new Hashtable<Description, Boolean>();

    /**
     * For testing (e.g. to check the past) you can set the today's date.
     *
     * @param date the new today's date
     */
    public void setToday(final Date date) {
        this.today = date;
    }

    /**
     * For testing you can set the OS name.
     *
     * @param name the new os name
     */
    public void setOsName(final String name) {
        this.osName = name;
    }

    /**
     * For testing you can set the user name.
     *
     * @param name the new user name
     */
    public void setUserName(final String name) {
        this.userName = name;
    }

    /**
     * Describes the kind of test e.g. "IntegrationTest".
     *
     * @return e.g. "all tests including integration tests"
     * @see org.junit.runner.manipulation.Filter#describe()
     */
    @Override
    public String describe() {
        if (Environment.isPropertyEnabled(Environment.INTEGRATION_TEST)) {
            return "all tests including integration tests";
        }
        if (Environment.isPropertyEnabled(Environment.RUN_SMOKE_TESTS)) {
            return "tests marked as @SmokeTest";
        }
        return "all tests (except tests marked as @Broken)";
    }

    /**
     * Handles the different annotations like Broken, RunTestOn, SkipTestOn,
     * IntegrationTest or SmokeTest.
     * We use a cache (shouldRunCache) here in this implementation to avoid to
     * much logging of "...run method / skip method..." messages.
     *
     * @param description containing the JUnit class and test method
     * @return depends on the annotation
     * @see org.junit.runner.manipulation.Filter#shouldRun(org.junit.runner.Description)
     */
    @Override
    public boolean shouldRun(final Description description) {
        Boolean value = shouldRunCache.get(description);
        if (value != null) {
            return value;
        }
        try {
            if (isSkipTestOn(description) || !isRunTestOn(description) || isBroken(description)) {
                value = false;
            } else if (Environment.smokeTestEnabled) {
                value = isSmokeTest(description);
            } else if (Environment.integrationTestEnabled) {
                value = true;
            } else {
                value = !isIntegrationTest(description);
            }
            shouldRunCache.put(description, value);
            if (!value) {
                filteredOut++;
            }
            return value;
        } catch (IllegalArgumentException iae) {
            log.warn(description + " has incomplete Annotation(s) - " + iae);
            return false;
        }
    }

    private boolean isIntegrationTest(final Description description) {
        IntegrationTest it = description.getAnnotation(IntegrationTest.class);
        if (it != null) {
            if (log.isInfoEnabled()) {
                log.info(description.getDisplayName() + " SKIPPED because "
                        + it.value());
            }
            return true;
        }
        return false;
    }

    private boolean isSmokeTest(final Description description) {
        SmokeTest smokeTest = description.getAnnotation(SmokeTest.class);
        if (smokeTest != null) {
            if (log.isInfoEnabled()) {
                log.info("run " + description.getDisplayName() + " because "
                        + smokeTest.value());
            }
            return true;
        }
        return false;
    }

    /**
     * Returns true if the given method has the RunTestOn annotation and
     * the condition of it is true.
     *
     * @param description with the FrameworkMethod
     * @return depends on the condition
     */
    private static boolean isRunTestOn(final Description description) {
        RunTestOn runOn = description.getAnnotation(RunTestOn.class);
        if (runOn == null) {
            return true;
        }
        TestOn testOn = new TestOn();
        if (testOn.matches(runOn.value(), runOn.osName(), runOn.osArch(), runOn
                .osVersion(), runOn.host(), runOn.javaVersion(), runOn
                .javaVendor(), runOn.user(), runOn.property())) {
            log.info(description + " executed " + testOn.getReason());
            return true;
        }
        return false;
    }

    /**
     * Returns true if the given method has the SkipTestOn annotation and
     * the condition of it is true.
     *
     * @param description with the FrameworkMethod
     * @return depends on the condition
     */
    private static boolean isSkipTestOn(final Description description) {
        SkipTestOn skipOn = description.getAnnotation(SkipTestOn.class);
        if (skipOn == null) {
            return false;
        }
        TestOn testOn = new TestOn();
        if (testOn.matches(skipOn.value(), skipOn.osName(), skipOn.osArch(), skipOn
                .osVersion(), skipOn.host(), skipOn.javaVersion(), skipOn
                .javaVendor(), skipOn.user(), skipOn.property())) {
            log.info(description + " SKIPPED because " + testOn.getReason());
            return true;
        }
        return false;
    }

    private boolean isBroken(final Description description) {
        Broken broken = description.getAnnotation(Broken.class);
        if (broken == null) {
            return false;
        }
        return isBroken(description.getMethodName(), broken);
    }


    /**
     * Checks for a given method if the condition of the given Broken
     * annotations is fulfilled.
     * <br/>
     * NOTE: The method has default visibility
     *
     * @param method the method
     * @param broken the broken
     * @return true, if is broken
     */
    @SuppressWarnings("deprecation")
    public boolean isBroken(final String method, final Broken broken) {
        String why = broken.why();
        if (StringUtils.isEmpty(why)) {
            why = broken.value();
        }
        Date till = Converter.toDate(broken.till());
        if ((till != NullConstants.NULL_DATE) && till.after(today)) {
            log.info(method + "() SKIPPED because " + broken.value()
                    + " till " + broken.till());
            return true;
        }
        if (StringUtils.isNotEmpty(broken.os()[0])) {
            log.warn("@Broken(os=...) is deprecated - use @Broken(osName=...)");
            if (StringHelper.containsIgnoreCase(broken.os(), osName)) {
                log.info(method + "() SKIPPED because " + why + " on "
                        + Converter.toString(broken.os()));
                return true;
            } else {
                log.debug(method + "() started on " + osName + ", because " + why + " only on "
                        + Converter.toString(broken.os()));
                return false;
            }
        }
        TestOn testOn = new TestOn();
        testOn.setOsName(this.osName);
        testOn.setUserName(this.userName);
        try {
            if (testOn.matches(broken.osName(), broken.osArch(), broken
                    .osVersion(), broken.host(), broken.javaVersion(), broken
                    .javaVendor(), broken.user(), broken.property())) {
                log.info(method + " SKIPPED because " + testOn.getReason());
                return true;
            }
        } catch (IllegalArgumentException canhappen) {
            log.trace(canhappen);
            if (till == NullConstants.NULL_DATE) {
                log.info(method + "() SKIPPED because " + why);
                return true;
            } else {
                log.debug(method + "() started, because it should be fixed since "
                        + broken.till());
                return false;
            }
        }
        return false;
    }

    /**
     * Checks if there are some methods filtered out.
     * @return true if some methods were filtered
     */
    public boolean hasFiltered() {
        return filteredOut > 0;
    }

    /**
     * Prints the description of this filter.
     *
     * @return a String including the result of the describe() method
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        return "filter for " + this.describe();
    }

}

