/*
 * $Id: SmokeRunner.java,v 1.15 2010/12/31 14:40:13 oboehm Exp $
 *
 * Copyright (c) 2010 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * (c)reated 29.03.2010 by oliver (ob@oasd.de)
 */

package patterntesting.runtime.junit;

import java.lang.annotation.Annotation;
import java.lang.reflect.*;
import java.util.*;


import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.*;
import org.junit.*;
import org.junit.internal.AssumptionViolatedException;
import org.junit.internal.runners.model.*;
import org.junit.internal.runners.statements.*;
import org.junit.runner.Description;
import org.junit.runner.manipulation.NoTestsRemainException;
import org.junit.runner.notification.*;
import org.junit.runners.*;
import org.junit.runners.model.*;

import patterntesting.runtime.annotation.*;
import patterntesting.runtime.junit.internal.ProfiledStatement;
import patterntesting.runtime.util.Environment;

/**
 * This is the eXtended Runner for JUnit 4 which handles the SmokeTest and
 * other annotations.
 * It can be used together with the {@code @RunWith} annotation of JUnit 4.
 *
 * @author oliver
 * @since 1.0 (29.03.2010)
 */
public class SmokeRunner extends ParentRunner<FrameworkMethod> {
//public class SmokeRunner extends BlockJUnit4ClassRunner {

    private static final Log log = LogFactory.getLog(SmokeRunner.class);
    private static final boolean runSmokeTests = Environment
            .isPropertyEnabled(Environment.RUN_SMOKE_TESTS);
    private final SmokeFilter xfilter;

    /**
     * Creates a SmokeRunner to run klass methods in parallel.
     *
     * @param klass the test class to run
     * @throws InitializationError if the test class is malformed
     */
    public SmokeRunner(final Class<?> klass) throws InitializationError {
        super(klass);
        xfilter = new SmokeFilter();
        try {
            this.filter(xfilter);
        } catch (NoTestsRemainException e) {
            if (xfilter.hasFiltered()) {
                log.info("all tests filtered out by " + xfilter);
            } else {
                throw new InitializationError(e);
            }
        }
    }

    /**
     * The ParentRunner allows us no access to the filter. So we added this
     * method here (e.g. needed by the ParallelRunner).
     *
     * @return the SmokeFilter
     */
    protected final SmokeFilter getFilter() {
        return this.xfilter;
    }

//    /**
//     * If we extends from BlockJUnit4CLassRunner we must overwrite this method
//     * here because we also support JUnit3 tests. Otherwise we would get an
//     * InitializationExcption.
//     */
//    @Override
//    protected void collectInitializationErrors(List<Throwable> errors) {
//    }

    /**
     * Returns the methods that run tests. Default implementation
     * returns all methods annotated with {@code @Test} on this
     * class and superclasses that are not overridden.
     *
     * @return the methods that run tests
     */
    @Override
    protected final List<FrameworkMethod> getChildren() {
        TestClass testClass = this.getTestClass();
        Class<?> javaClass = testClass.getJavaClass();
        if (ProfiledStatement.isTestCaseClass(testClass)) {
            return getJUnit3TestMethods(javaClass);
        }
        return testClass.getAnnotatedMethods(Test.class);
    }

    /**
     * Create a <code>Description</code> of a single test named <code>name</code> in the class <code>clazz</code>.
     * Generally, this will be a leaf <code>Description</code>.
     * (see also {@link BlockJUnit4ClassRunner})
     *
     * @param child the name of the test (a method name for test annotated with {@link org.junit.Test})
     * @see org.junit.runners.ParentRunner#describeChild(java.lang.Object)
     */
    @Override
    protected final Description describeChild(final FrameworkMethod child) {
        return Description.createTestDescription(getTestClass().getJavaClass(),
                child.getName(), child.getAnnotations());
    }

    private final Description describeTestClass() {
        return Description.createSuiteDescription(getTestClass().getName(), getTestClass()
                .getAnnotations());
    }

    /**
     * Checks the annotation of the method marked as "@BeforeClass" and add
     * (or filters out) the beforeClass method (needed to solve
     * {@link "https://sourceforge.net/tracker/?func=detail&aid=3034823&group_id=48833&atid=454317"}).
     *
     * @see org.junit.runners.ParentRunner#withBeforeClasses(org.junit.runners.model.Statement)
     * @param statement the statement(s) before
     * @return the statement(s) after with the BeforeClass method
     */
    @Override
    protected Statement withBeforeClasses(final Statement statement) {
        List<FrameworkMethod> filtered = filter(BeforeClass.class);
        if (filtered.isEmpty()) {
            return statement;
        }
        return new RunBefores(statement, filtered, null);
    }

    /**
     * Checks the annotation of the method marked as "@AfterClass" and add
     * (or filters out) the afterClass method (needed to solve
     * {@link "https://sourceforge.net/tracker/?func=detail&aid=3034823&group_id=48833&atid=454317"}).
     *
     * @see org.junit.runners.ParentRunner#withAfterClasses(org.junit.runners.model.Statement)
     * @param statement the statement(s) before
     * @return the statement(s) after with the AfterClass method
     */
    @Override
    protected Statement withAfterClasses(final Statement statement) {
        List<FrameworkMethod> filtered = filter(AfterClass.class);
        if (filtered.isEmpty()) {
            return statement;
        }
        return new RunAfters(statement, filtered, null);
    }

    private List<FrameworkMethod> filter(final Class<? extends Annotation> annotationClass) {
        List<FrameworkMethod> methods = this.getTestClass().getAnnotatedMethods(annotationClass);
        return filter(methods);
    }

    private List<FrameworkMethod> filter(final List<FrameworkMethod> methods) {
        List<FrameworkMethod> filtered = new ArrayList<FrameworkMethod>();
        for (FrameworkMethod fm : methods) {
            if (this.xfilter.shouldRun(describeChild(fm))) {
                filtered.add(fm);
            } else if (log.isTraceEnabled()) {
                log.trace(fm.getMethod() + " is filtered out");
            }
        }
        return filtered;
    }


    /**
     * Checks if the whole class can be ignored.
     *
     * @param notifier the notifier
     * @see org.junit.runners.ParentRunner#run(org.junit.runner.notification.RunNotifier)
     */
    @Override
    public final void run(final RunNotifier notifier) {
        if (this.shouldBeIgnored()) {
            EachTestNotifier testNotifier = new EachTestNotifier(notifier,
                    getDescription());
            testNotifier.fireTestIgnored();
            testNotifier.fireTestFinished();
        } else {
            super.run(notifier);
        }
    }

    /**
     * If some tests are ignored we must correct the number.
     * TODO use Filter class for filtering
     *
     * @return number of test methods
     * @see org.junit.runner.Runner#testCount()
     */
    @Override
    public final int testCount() {
        if (this.shouldBeIgnored()) {
            return 0;
        }
        return super.testCount();
    }

    /**
     * Runs the test corresponding to {@code child}, which can be assumed to be
     * an element of the list returned by {@link ParentRunner#getChildren()}.
     * Subclasses are responsible for making sure that relevant test events are
     * reported through {@code notifier}
     */
    @Override
    protected final void runChild(final FrameworkMethod method,
            final RunNotifier notifier) {
        Description description = describeChild(method);
        notifier.fireTestStarted(description);
        try {
            if (shouldBeIgnored(method) || !shouldBeRun(method)) {
                notifier.fireTestIgnored(description);
                return;
            }
        } catch (IllegalArgumentException iae) {
            notifier.fireTestAssumptionFailed(new Failure(description, iae));
            notifier.fireTestFinished(description);
            return;
        }
        Statement stmt = methodBlock(method);
        try {
            stmt.evaluate();
        } catch (AssumptionViolatedException e) {
            notifier.fireTestAssumptionFailed(new Failure(description, e));
        } catch (Throwable e) {
            addFailure(notifier, e, description);
        } finally {
            logStatement(stmt);
            notifier.fireTestFinished(description);
        }
    }

    /**
     * We will give a subclass (like e.g. ParallelRunner) the chance to report
     * the statement with its own logger.
     *
     * @param stmt the stmt to be logged
     */
    protected void logStatement(final Statement stmt) {
        log.info(stmt);
    }

    /**
     * Here we handle the annotations like {@code @TestIntegration} or
     * {@code @Broken} for the whole class.
     *
     * @return true if test should be ignored
     */
    private boolean shouldBeIgnored() {
        TestClass testClass = this.getTestClass();
        if (runSmokeTests) {
            return !hasSmokeTests(testClass);
        }
        Annotation[] annotations = testClass.getAnnotations();
        for (int i = 0; i < annotations.length; i++) {
            Annotation a = annotations[i];
            if (a.annotationType().equals(IntegrationTest.class)
                    && isIntegrationTest((IntegrationTest) a)) {
                return true;
            }
        }
        return !this.xfilter.shouldRun(describeTestClass());
//        return false;
    }

    /**
     * Here we handle annotations like {@code @Ignore} where the execution of
     * the test method should be ignored.
     *
     * @TODO should be handled by XFilter class
     *
     * @param method the test method
     * @return true or false
     */
    protected final boolean shouldBeIgnored(final FrameworkMethod method) {
        Ignore ignore = method.getAnnotation(Ignore.class);
        if (ignore != null) {
            if (log.isDebugEnabled()) {
                String reason = ignore.value();
                if (StringUtils.isNotEmpty(reason)) {
                    reason = " (" + reason + ")";
                }
                log.debug(this.getTestClass().getName() + "."
                        + method.getName() + " ignored" + reason);
            }
            return true;
        }
        return false;
    }

    private boolean shouldBeRun(final FrameworkMethod method) {
        return this.xfilter.shouldRun(describeChild(method));
    }

    /**
     * This method was inspired from an internal JUnit class
     * (EachTestNotifier#addFailure(Throwable)).
     */
    private void addFailure(final RunNotifier notifier, final Throwable targetException, final Description description) {
        if (targetException instanceof MultipleFailureException) {
            MultipleFailureException mfe = (MultipleFailureException) targetException;
            for (Throwable each : mfe.getFailures()) {
                addFailure(notifier, each, description);
            }
            return;
        }
        notifier.fireTestFailure(new Failure(description, targetException));
    }

    /**
     * Creates a RunStatement for the given test method.
     *
     * @param method the test method
     * @return a created RunStatement
     */
    protected Statement methodBlock(final FrameworkMethod method) {
        return new ProfiledStatement(this.getTestClass(), method);
    }

    /**
     * Here we look after public void methods with "test" as prefix and with no
     * arguments.
     * <br/>
     * NOTE: This method is public because it is also needed by
     * patterntesting.concurrent.junit.JUnit3Executor
     *
     * @param testClass the test class
     * @return a list of public methods starting with prefix "test"
     */
    public static List<FrameworkMethod> getJUnit3TestMethods(final Class<?> testClass) {
        List<FrameworkMethod> children = new ArrayList<FrameworkMethod>();
        Method[] methods = testClass.getDeclaredMethods();
        for (int i = 0; i < methods.length; i++) {
            Method method = methods[i];
            int mod = method.getModifiers();
            if (Modifier.isPrivate(mod) || Modifier.isStatic(mod)) {
                continue;
            }
            if (method.getParameterTypes().length > 0) {
                continue;
            }
            Class<?> returnType = method.getReturnType();
            if (!returnType.toString().equalsIgnoreCase("void")) {
                continue;
            }
            String name = method.getName();
            if (name.startsWith("test")) {
                if (Modifier.isPublic(mod)) {
                    FrameworkMethod child = new FrameworkMethod(method);
                    children.add(child);
                } else {
                    log.warn(method + " isn't public");
                }
            }
        }
        return children;
    }



    /////   Annotation Handling   /////////////////////////////////////////////

    private boolean isIntegrationTest(final IntegrationTest it) {
        if (Environment.integrationTestEnabled) {
            log.trace("all tests including integration tests will be executed");
            return false;
        }
        log.info(this.getTestClass().getName() + " SKIPPED because "
                + it.value() + " (can be activated with '-D"
                + Environment.INTEGRATION_TEST + "')");
        return true;
    }

    /**
     * Looks for the given test class if it has the SmokeClass annotation or
     * one of its test methods have it.
     *
     * @param testClass the test class
     * @return true if smoke tests are found
     */
    private static boolean hasSmokeTests(final TestClass testClass) {
        if (testClass.getJavaClass().getAnnotation(SmokeTest.class) != null) {
            return true;
        }
        List<FrameworkMethod> smokeTests = testClass
                .getAnnotatedMethods(SmokeTest.class);
        return !smokeTests.isEmpty();
    }

}

