/*
 * $Id: ReflectionHelper.java,v 1.5 2010/12/30 17:33:23 oboehm Exp $
 *
 * Copyright (c) 2009 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * (c)reated 09.03.2009 by oliver (ob@aosd.de)
 */
package patterntesting.runtime.util;

import java.lang.reflect.*;
import java.util.*;

import org.apache.commons.logging.*;

/**
 * This class is a helper class to access some fields via reflection. Normally
 * you should avoid reflection. Handle with care if you use it.
 *
 * @author <a href="boehm@javatux.de">oliver</a>
 * @since 09.03.2009
 * @version $Revision: 1.5 $
 */
public class ReflectionHelper {

     private static final Log log = LogFactory.getLog(ReflectionHelper.class);

     /** Utility class - no need to instantiate it */
     private ReflectionHelper() {}

    /**
     * Tries to get the wanted field. Before it is returned the acc
     *
     * @param cl the cl
     * @param name the name
     *
     * @return the wanted field
     *
     * @throws NoSuchFieldException the no such field exception
     */
    public static Field getField(final Class<?> cl, final String name)
            throws NoSuchFieldException {
        try {
            Field field = cl.getDeclaredField(name);
            field.setAccessible(true);
            return field;
        } catch (NoSuchFieldException e) {
            Class<?> superclass = cl.getSuperclass();
            if (superclass == null) {
                throw e;
            } else {
                if (log.isTraceEnabled()) {
                    log.trace("using " + superclass + " to get " + cl.getName()
                            + "." + name + "...");
                }
                return ReflectionHelper.getField(superclass, name);
            }
        }
    }

    /**
     * If it can it returns the value of the given field. If not it throws an
     * exception.
     *
     * @param obj the obj
     * @param name the name
     *
     * @return the field value
     *
     * @throws NoSuchFieldException the no such field exception
     * @throws IllegalAccessException the illegal access exception
     * @throws IllegalArgumentException the illegal argument exception
     */
    public static Object getFieldValue(final Object obj, final String name)
            throws NoSuchFieldException, IllegalArgumentException,
            IllegalAccessException {
        Field field = ReflectionHelper.getField(obj.getClass(), name);
        return field.get(obj);
    }

    /**
     * To get all uninitialized field you can call this method.
     *
     * @param obj the object where to get the fields from
     *
     * @return a collecion of unitialized fields (can be empty)
     */
    public static Collection<Field> getUninitializedNonStaticFields(final Object obj) {
        Collection<Field> unitializedFields = new ArrayList<Field>();
        Field[] fields = obj.getClass().getDeclaredFields();
        for (int i = 0; i < fields.length; i++) {
            try {
                fields[i].setAccessible(true);
                if ((fields[i].get(obj) == null) && !isStatic(fields[i])) {
                    unitializedFields.add(fields[i]);
                }
            } catch (IllegalArgumentException e) {
                log.info(e + " => " + fields[i] + " ignored");
            } catch (IllegalAccessException e) {
                log.debug("can't access " + fields[i] + " => ignored");
            }
        }
        return unitializedFields;
    }

    /**
     * Checks if the given field is static.
     *
     * @param field the field
     * @return true, if is static
     */
    public static boolean isStatic(final Field field) {
        int m = field.getModifiers();
        return Modifier.isStatic(m);
    }

    /**
     * To short string.
     *
     * @param field the field
     *
     * @return the string
     */
    public static String toShortString(final Field field) {
        return field.getType().getSimpleName() + " " + field.getName();
    }

    /**
     * To short string.
     *
     * @param fields the fields
     *
     * @return the string
     */
    public static String toShortString(final Collection<Field> fields) {
        StringBuffer sbuf = new StringBuffer();
        for (Iterator<Field> iterator = fields.iterator(); iterator.hasNext();) {
            Field field = iterator.next();
            sbuf.append(", ");
            sbuf.append(toShortString(field));
        }
        return sbuf.substring(2);
    }

    /**
     * Gets the method.
     *
     * @param cl e.g. class java.lang.ClassLoader
     * @param name e.g. "getPackages"
     *
     * @return a Method object
     *
     * @throws NoSuchMethodException the no such method exception
     */
	public static Method getMethod(final Class<?> cl, final String name) throws NoSuchMethodException {
	    try {
	        return cl.getDeclaredMethod(name);
	    } catch (NoSuchMethodException e) {
	        Class<?> superclass = cl.getSuperclass();
	        if (superclass == null) {
	            throw e;
	        } else {
	            return getMethod(superclass, name);
	        }
	    }
	}

}
