/*
 * $Id: SignatureHelper.java,v 1.4 2010/12/30 17:33:23 oboehm Exp $
 *
 * Copyright (c) 2009 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * (c)reated 03.04.2009 by oliver (ob@aosd.de)
 */
package patterntesting.runtime.util;

import java.lang.reflect.*;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.*;
import org.aspectj.lang.Signature;
import org.aspectj.lang.reflect.*;

import patterntesting.runtime.util.reflect.*;

/**
 * The Class SignatureHelper.
 *
 * @author <a href="boehm@javatux.de">oliver</a>
 * @since 03.04.2009
 * @version $Revision: 1.4 $
 */
public class SignatureHelper {

    private static final Log log = LogFactory.getLog(SignatureHelper.class);

    /** Utility class - no need to instantiate it */
    private SignatureHelper() {}

    /**
     * The difference to Signature.toString() is that in case of a methode
     * or constructor the return type is not part of result.
     *
     * @param sig the sig
     *
     * @return the given signature as string
     */
    public static String getAsString(final Signature sig) {
        try {
            CodeSignature codeSig = (CodeSignature) sig;
            String name = getAsString(sig.getDeclaringTypeName(), sig);
            Class<?>[] params = codeSig.getParameterTypes();
            return name + "(" + Converter.toShortString(params) + ")";
        } catch (Exception e) {
            log.debug("can't get '" + sig + "' as String", e);
            return sig.toString();
        }
    }

    /**
     * Gets the as string.
     *
     * @param type the type
     * @param sig the sig
     *
     * @return the as string
     */
    public static String getAsString(final String type, final Signature sig) {
        String name = sig.getName();
        if (name.equals("<init>")) {
            return "new " + type;
        } else {
            return type + "." + name;
        }
    }

    /**
     * Gets the as signature.
     *
     * @param method the method
     *
     * @return the as signature
     */
    public static MethodSignature getAsSignature(final Method method) {
        return new MethodSignatureImpl(method);
    }

    /**
     * Gets the as signature.
     *
     * @param ctor the ctor
     *
     * @return the as signature
     */
    public static ConstructorSignature getAsSignature(final Constructor<?> ctor) {
        return new ConstructorSignatureImpl(ctor);
    }

    /**
     * Gets the as signature.
     *
     * @param label e.g. "java.lang.String.substring(int)" or
     * "new java.lang.String()
     *
     * @return the given label as Signature
     *
     * @throws ClassNotFoundException the class not found exception
     * @throws NoSuchMethodException the no such method exception
     * @throws SecurityException the security exception
     */
    public static Signature getAsSignature(final String label)
            throws ClassNotFoundException, SecurityException,
            NoSuchMethodException {
        if (label.startsWith("new ")) {
            return getAsConstructorSignature(label);
        }
        String classDotMethod = StringUtils.substringBefore(label, "(");
        String className = StringUtils.substringBeforeLast(classDotMethod, ".");
        String methodName = StringUtils.substringAfterLast(classDotMethod, ".");
        Class<?>[] params = getParameterTypes(label);
        return getAsSignature(Class.forName(className), methodName, params);
    }

    private static Signature getAsSignature(final Class<?> clazz, final String methodName,
            final Class<?>[] parameterTypes) throws ClassNotFoundException, SecurityException,
            NoSuchMethodException {
        Method method = clazz.getMethod(methodName, parameterTypes);
        return getAsSignature(method);
    }

    private static ConstructorSignature getAsConstructorSignature(final String label)
            throws ClassNotFoundException, SecurityException,
            NoSuchMethodException {
        String ctor = StringUtils.substringAfterLast(label, " ").trim();
        String className = StringUtils.substringBefore(ctor, "(");
        Class<?>[] params = getParameterTypes(label);
        return getAsConstructorSignature(Class.forName(className), params);
    }

    private static ConstructorSignature getAsConstructorSignature(final Class<?> clazz,
            final Class<?>[] parameterTypes) throws ClassNotFoundException, SecurityException,
            NoSuchMethodException {
        Constructor<?> ctor = clazz.getConstructor(parameterTypes);
        return getAsSignature(ctor);
    }

    private static Class<?>[] getParameterTypes(final String label)
            throws ClassNotFoundException {
        String parameters = StringUtils.substringAfter(label, "(");
        parameters = StringUtils.substringBefore(parameters, ")");
        String[] params = StringUtils.split(parameters, ',');
        Class<?>[] parameterTypes = new Class<?>[params.length];
        for (int i = 0; i < params.length; i++) {
            parameterTypes[i] = Class.forName(params[i].trim());
        }
        return parameterTypes;
    }

}

/**
 * $Log: SignatureHelper.java,v $
 * Revision 1.4  2010/12/30 17:33:23  oboehm
 * checkstyle warnings reduced
 *
 * Revision 1.3  2010/04/22 18:32:01  oboehm
 * compiler warnings fixed
 *
 * Revision 1.2  2010/04/22 18:27:19  oboehm
 * code cleanup of src/main/java
 *
 * Revision 1.1  2010/01/05 13:26:17  oboehm
 * begin with 1.0
 *
 * Revision 1.6  2009/12/19 22:34:09  oboehm
 * trailing spaces removed
 *
 * Revision 1.5  2009/09/25 14:49:43  oboehm
 * javadocs completed with the help of JAutodoc
 *
 * Revision 1.4  2009/09/18 13:54:52  oboehm
 * javadoc warnings fixed
 *
 * Revision 1.3  2009/04/09 18:30:47  oboehm
 * no longer double entries (bug 2745302 fixed)
 *
 * $Source: /cvsroot/patterntesting/PatternTesting10/patterntesting-rt/src/main/java/patterntesting/runtime/util/SignatureHelper.java,v $
 */
